import math

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["Identity", "Flatten", "Linear", "Bilinear"]


class Identity(nn.Module):
    def forward(self, input_: Tensor) -> Tensor:
        return input_


class Flatten(nn.Module):
    def __init__(self, start_axis: int = 0, end_axis: int = -1) -> None:
        super().__init__()
        self.start_axis = start_axis
        self.end_axis = end_axis

    def forward(self, input_: Tensor) -> Tensor:
        return lucid.flatten(input_, self.start_axis, self.end_axis)


class Linear(nn.Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

        weight_ = lucid.empty(out_features, in_features)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.empty(out_features)
            self.bias = nn.Parameter(bias_)
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.kaiming_uniform(self.weight)

        if self.bias is not None:
            fan_in, _ = nn.init._dist._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform(self.bias, -bound, bound)

    def forward(self, input_: Tensor) -> Tensor:
        return F.linear(input_, self.weight, self.bias)

    def extra_repr(self) -> str:
        return f"{self.in_features}, {self.out_features}, bias={self.bias is not None}"


class Bilinear(nn.Module):
    def __init__(
        self, in1_features: int, in2_features: int, out_features: int, bias: bool = True
    ) -> None:
        super().__init__()
        self.in1_features = in1_features
        self.in2_features = in2_features
        self.out_features = out_features

        weight_ = lucid.empty(out_features, in1_features, in2_features)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.empty(out_features)
            self.bias = nn.Parameter(bias_)
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        bound = 1 / math.sqrt(self.weight.shape[1])
        nn.init.uniform(self.weight, -bound, bound)

        if self.bias is not None:
            nn.init.uniform(self.bias, -bound, bound)

    def forward(self, input_1: Tensor, input_2: Tensor) -> Tensor:
        return F.bilinear(input_1, input_2, self.weight, self.bias)

    def extra_repr(self) -> str:
        return (
            f"{self.in1_features}, {self.in2_features}, {self.out_features}, "
            f"bias={self.bias is not None}"
        )



from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.objdet.util import iou


__all__ = ["YOLO_V1"]


arch_config = [
    (64, 7, 2, 3),
    "M",
    (192, 3, 1, 1),
    "M",
    (128, 1, 1, 0),
    (256, 3, 1, 1),
    (256, 1, 1, 0),
    (512, 3, 1, 1),
    "M",
    [(256, 1, 1, 0), (512, 3, 1, 1), 4],
    (512, 1, 1, 0),
    (1024, 3, 1, 1),
    "M",
    [(512, 1, 1, 0), (1024, 3, 1, 1), 2],
    (1024, 3, 1, 1),
    (1024, 3, 2, 1),
    (1024, 3, 1, 1),
    (1024, 3, 1, 1),
]


class _ConvBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, **kwargs) -> None:
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels)
        self.act = nn.LeakyReLU(0.1)

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.bn(self.conv(x)))


class YOLO_V1(nn.Module):
    def __init__(
        self,
        in_channels: int,
        split_size: int,
        num_boxes: int,
        num_classes: int,
        lambda_obj: float = 5.0,
        lambda_noobj: float = 5.0,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.split_size = split_size
        self.num_boxes = num_boxes
        self.num_classes = num_classes
        self.lambda_obj = lambda_obj
        self.lambda_noobj = lambda_noobj

        self.darknet = self._create_conv_layers(arch=arch_config)
        self.fcs = nn.Sequential(
            nn.Linear(1024 * split_size**2, 4096),
            nn.LeakyReLU(0.1),
            nn.Linear(4096, split_size**2 * (num_boxes * 5 + num_classes)),
        )

    def _create_conv_layers(self, arch: list) -> nn.Sequential:
        layers = []
        in_channels = self.in_channels
        for cfg in arch:
            if isinstance(cfg, tuple):
                layers.append(
                    _ConvBlock(
                        in_channels,
                        cfg[0],
                        kernel_size=cfg[1],
                        stride=cfg[2],
                        padding=cfg[3],
                    )
                )
                in_channels = cfg[0]

            elif isinstance(cfg, str):
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))

            elif isinstance(cfg, list):
                conv1, conv2, num_repeats = cfg
                for _ in range(num_repeats):
                    layers.append(
                        _ConvBlock(
                            in_channels,
                            conv1[0],
                            kernel_size=conv1[1],
                            stride=conv1[2],
                            padding=conv1[3],
                        )
                    )
                    layers.append(
                        _ConvBlock(
                            conv1[0],
                            conv2[0],
                            kernel_size=conv2[1],
                            stride=conv2[2],
                            padding=conv2[3],
                        )
                    )
                    in_channels = conv2[0]

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.darknet(x)
        x = x.flatten(start_axis=1)
        x = self.fcs(x)
        return x

    def get_loss(self, x: Tensor, target: Tensor) -> Tensor:
        N = x.shape[0]
        S, B, C = self.split_size, self.num_boxes, self.num_classes

        pred = self.forward(x).reshape(N, S, S, B * 5 + C)
        target = target.reshape(N, S, S, B * 5 + C)

        obj_mask = target[..., 4:5]
        noobj_mask = 1.0 - obj_mask

        pred_boxes = pred[..., : B * 5].reshape(N, S, S, B, 5)
        target_box = target[..., :5]

        pred_xy = pred_boxes[..., :2]
        pred_wh = pred_boxes[..., 2:4] / 2
        pred_corners = lucid.concatenate(
            [pred_xy - pred_wh, pred_xy + pred_wh], axis=-1
        )

        target_xy = target_box[..., 2:]
        target_wh = target_box[..., 2:4] / 2
        tatget_corners = lucid.concatenate(
            [target_xy - target_wh, target_xy + target_wh], axis=-1
        )

        ious = []
        for b in range(B):
            iou_val = iou(
                pred_corners[..., b, :].reshape(-1, 4),
                tatget_corners.reshape(-1, 4),
            )
            # TODO: implement `lucid.Tensor.diagonal`
